/* START OF user_config.c ------------------------------------------------------
 *
 * 	Locate/create user configuration files.
 *
 * 	There are two versions of the below, a UNIX version (WITH_UNIX_EXTRAS)
 *	and an ANSI C version.  The ANSI C version will fail (#error) where you
 *	need to add OS specific code.
 *
 *	(It'll only support UNIX and <other>, not multiple systems, rethink it.)
 *
 *	(This again is not fantastic code, but if I'm going to get anywhere with
 *	this I'm going to have to speedup development a bit.)
 *
 *      ---
 *      Written by David Eggleston (2006) <deggleston@users.sourceforge.net>
 *      for the 'BeebEm' Acorn BBC Model B, Integra-B, Model B Plus and
 *      Master 128 emulator.
 *
 *      This file is part of BeebEm and may be copied only under the terms of
 *	either the GNU General Public License (GPL) or Dr. David Alan Gilbert's
 *	BeebEm license.
 *
 *      For more details please visit:
 *
 *      http://www.gnu.org/copyleft/gpl.html
 *      ---
 */

#if HAVE_CONFIG_H
#       include <config.h>
#endif

#include "include/gui_functions.h"
#include "include/user_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Extra UNIX specific includes:
 */
#ifdef WITH_UNIX_EXTRAS
#	include <sys/types.h>
#	include <unistd.h>
#	include <pwd.h>
#	include <errno.h>
#	include <sys/stat.h>
#	include <dirent.h>
#endif



/* Sub directory (from DATA_DIR) for user configuration directory when not UNIX.
 */
#define NONE_UNIX_CONFIG_DIR "config"


/* Max length a path (including filename) can be for target OS.
 */
#define MAX_PATH_LEN	(1024*4+1)

/* Filename and path manipulation functions, not too formal, but does handle
 * more obvious errors.
 */
static BOOL AppendStringToPath(char *path, char *append, size_t len)
{
	/* Don't bother appending if NULL, but do warn.
	 */
	if (path == NULL){
		qERROR("Appending to NULL path.");
		return(FALSE);
	}
	if (append == NULL){
		qERROR("Appending NULL to path.");
		return(FALSE);
	}

	/* Will the characters to append fit?
	 */
	if (strlen(path)+strlen(append) < len)
		/* Yes, so append it.
		 */
		strcat(path, append);
	else{
		qERROR("Path too long.");
		return(FALSE);
	}

	return(TRUE);
}

/* Will append a forward slash to the path if one does not already exist
 */
static BOOL AppendForwardSlashToPath(char *path, size_t len)
{
	/* Don't bother if path NULL.
	 */
	if (path == NULL){
		qERROR("Appending directory to NULL path.");
		return(FALSE);
	}

	if (strlen(path)>0 && path[strlen(path)-1] != '/')
		return( AppendStringToPath(path, "/", len) );

	return(TRUE);
}

/* Will append a directory name to a path (and handle forward slash).  The
 * directory name should not contain slashes.
 */
static BOOL AppendDirToPath(char *path, char *dir, size_t len)
{
	/* Don't bother if NULL.
	 */
	if (path == NULL){
		qERROR("Appending directory to NULL path.");
		return(FALSE);
	}
	if (dir == NULL){
		qERROR("Appending NULL directory to path.");
		return(FALSE);
	}

	/* Append "/" to path if needed.
	 */
	if (! AppendForwardSlashToPath(path, len) )
		return(FALSE);

	/* Append Directory to Path/
	 */
	if (! AppendStringToPath(path, dir, len) )
		return(FALSE);

	/* As it's a dir, make sure a forward slash is appended too.
	 */
	if (! AppendForwardSlashToPath(path, len) )
		return(FALSE);

	return(TRUE);
}

/* Will append a filename to a path.
 */
static BOOL AppendFilenameToPath(char *path, char *filename, size_t len)
{
        /* Don't bother if NULL.
         */
        if (path == NULL){
                qERROR("Appending directory to NULL path.");
                return(FALSE);
        }
        if (filename == NULL){
                qERROR("Appending NULL filename to path.");
                return(FALSE);
        }

	/* Append "/" to path if needed.
	 */
	if (! AppendForwardSlashToPath(path, len) )
		return(FALSE);

	/* Append filename to Path
	 */
	if (! AppendStringToPath(path, filename, len) )
		return(FALSE);

	return(TRUE);
}

/*	*	*	*	*	*	*	*	*	*
 */


/* Test whether a regualr file exists.
 */
static BOOL TestFile(char *file)
{
	FILE *f;

	if ( (f=fopen(file, "r")) == NULL)
		return(FALSE);

	fclose(f);
	return(TRUE);
}


/* Test whether a directory entry exists.
 */
static BOOL TestDirectory(char *file)
{
#ifdef WITH_UNIX_EXTRAS
	DIR *dir;

	if ( (dir=opendir(file)) == NULL)
		return(FALSE);
	
	closedir(dir);
	return(TRUE);
#else
#	error You need to add TestDirectory support for your OS here.
#endif
}

/* Make a copy of a file.
 *
 * Both paths must be fully quantified (leading "/") for security reasons.
 */
static BOOL CopyFile(char *source_file, char *dest_file)
{
	/* Terminate program if file paths are not fully quantified.
	 *
	 * (They always should be for BeebEm, and if they're not, then it could
	 * be a security risk).
	 */
	if ( source_file == NULL || dest_file == NULL
	 || strlen(source_file) == 0 || strlen(dest_file) == 0
	 || source_file[0] != '/' || dest_file[0] != '/') {
		qFATAL("CopyFile arguments are not fully quantified"
		 " paths (bailing out).");
		exit(1);
	}

	/* Use 'cp' command instead of copying the files in C (cp must be
	 * in your PATH).
	 */
#ifdef WITH_SYSTEM_CP
	char command[5*1024];

	sprintf(command, "cp %s %s", source_file, dest_file);
	if (system(command) != 0){
		qERROR("cp command failed.");
		return(FALSE);
	}else
		return(TRUE);
#else
	/* Use C to copy the file instead of relying on 'cp'.
	 */
	FILE *src_f, *dst_f;
	int c, failed=0;

	if (source_file == NULL){
		qERROR("Source file is null string.");
		return FALSE;
	}
	if (dest_file == NULL){
		qERROR("Destination file is null string.");
		return FALSE;
	}

	/* Open the source file for reading, quit on fail.
	 */
	if ( (src_f=fopen(source_file, "r")) == NULL){
		pERROR(dL"Unable to open source file [%s]."
		 , dR, source_file);
		return FALSE;
	}

	/* Open and truncate the dest file for writing,
	 * close source file and quit on fail.
	 */
	if ( (dst_f=fopen(dest_file, "w")) == NULL){
		pERROR(dL"Unable to write to dest file [%s]."
		 , dR, dest_file);
		fclose(src_f);
		return FALSE;
	}

	/* Copy data from the source file to the dest file.
	 * (this is not fast!)
	 */
	while ( (c=fgetc(src_f)) != EOF){
		if (fputc(c, dst_f)==EOF){
			failed=1;
			break;
		}
	}

	/* Close files
	 */
	fclose(dst_f);
	fclose(src_f);

	if (failed==1){
		qERROR("File copy failed during copy.");
		return FALSE;
	}else
		return TRUE;
#endif
}


/* Create a new directory (succeeds if dir already exists).
 */
static BOOL CreateDirectory(char *file)
{
#ifdef WITH_UNIX_EXTRAS
//	printf("************** trying to create directory: %s\n", file);

	if (TestDirectory(file))
		return(TRUE);

	errno = 0;
	if (mkdir(file, (mode_t) S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP
	 | S_IXGRP | S_IROTH | S_IXOTH ) == 0)
		return(TRUE);

	pFATAL(dL"Could not create directory: %s", dR, strerror(errno));
	return(FALSE);
#else
#	error You need to add CreateDirectory support for your OS here.
#endif
}

/* Get users config directory ("${DATA_DIR}/config/" or "$HOME/.beebem/),
 * does add a trailing forward slash.
 *
 * On UNIX for a user called dave, would return a pointer to something like:
 * "/home/dave/.beebem/".  If this directory does not exist, it's created.
 *
 * On None UNIX systems, the installation path of the application + a config
 * directory name are used instead.  So something like:
 * "/applications/beebem/config/" will be returned instead. Of this directory
 * does not exist, it's created.
 */
char* GetUserConfigPath(char *path, size_t len)
{
	/* Return with error if string illegal length of 3 is valid.
	 */
	if (len<2) return NULL;

	/* Clear provided string
	 */
	path[0] = 0;

#ifdef WITH_UNIX_EXTRAS
	struct passwd *pw_ptr;

	/* Find user in password database.
	 */
	errno = 0; 
	if ( ! (pw_ptr=getpwuid(getuid())) ){
		/* Not found, clear config filename.
		 */
		pFATAL(dL"Unable to determine user home dir: %s", dR, strerror(errno));
		exit(1);
	}else{

		if (! AppendDirToPath(path, pw_ptr->pw_dir, len) ){
			qFATAL("Unable to accept user accounts dir.");
			exit(1);
		}

		if (! AppendDirToPath(path, "." PACKAGE, len) ){
			qFATAL("Unable to append ." PACKAGE " to user accounts dir.");
			exit(1);
		}
	}
#else
	/* Root of application data dir
	 */
	
	if (! AppendDirToPath(path, DATA_DIR, len) ){
		qFATAL("Unable to determine default user config dir.");
		exit(1);
	}

	/* Add config dir
	 */
	if (! AppendDirToPath(path,NONE_UNIX_CONFIG_DIR, len) ){
		qFATAL("Unable to accept default user config dir.");
		exit(1);
	}
#endif

	/* If configuration directory does not exist, create it.
	 */
	if (! TestDirectory(path) )
		if (! CreateDirectory(path)){
			qFATAL("Unable to create user config dir.");
			exit(1);
		}

	return(path);
}


/* The following functions are called within the emulator core to determine
 * the location of config files.
 *
 * If the file does not exist, most of these functions will try to create it
 * locally (in the users config directory) first.
 */

static BOOL MakeSureBufferIsSane(char *buffer, size_t length)
{
	/* Make sure buffer is sane.
	 */
	if (length<1){
		qERROR("Buffer is too small to contain a string.");
		return FALSE;
	}
	if (buffer == NULL){
		qERROR("Buffer is NULL.");
		return FALSE;
	}

	/* Make sure that whatever happens from now on the buffer
	 * is blank ("").
	 */
	buffer[0]=0;

	return TRUE;
}

static BOOL GetConfigFilePath(char *filepath, size_t length, char *file)
{
	/* Determine path (GetUserConfigPath will make sure
	 * the .beebem directory is created if it does not exist).
	 */
	if ( GetUserConfigPath(filepath, length) == NULL ){
		pERROR(dL"Unable to determine %s file path.", dR, file);
		return FALSE;
	}

	/* Append file to the path.
	 */
	if (! AppendFilenameToPath(filepath, file, length) ){
		return FALSE;
	}

	return TRUE;
}

static BOOL CreateFileFromMaster(char *source_filepath, char *master_filepath)
{
	/* Test filepath, if does not exist create it:
	 */
	if (TestFile(source_filepath) == FALSE){
		/* Make local copy (in ~/.beebem/ for UNIX)
		 */
		if (CopyFile(master_filepath, source_filepath) == FALSE){
			pERROR(dL"Unable to make copy of master file '%s'.", dR
			 , master_filepath);
			return FALSE;
		}
	}

	return TRUE;
}

static char* CopyFilePathToBuffer(char *filepath, char *buffer
 , size_t buffer_length)
{
	/* Copy filepath to buffer. Fail if buffer too small.
	 * Return pointer to buffer.
	 */
	if (strlen(filepath)+1 <= buffer_length){
		strcpy(buffer, filepath);
		return(buffer);
	}else
		return(NULL);
}

/* Return "~/.beebem/roms.cfg", if does not exist creates it using
 * master copy in "share/beebem/config/roms.cfg".
 */
char* GetLocation_roms_cfg(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	if (GetConfigFilePath(filepath, MAX_PATH_LEN, "roms.cfg") != TRUE)
		return NULL;

	if (CreateFileFromMaster(filepath, DATA_DIR"/config/roms.cfg") == FALSE)
		return NULL;

	return CopyFilePathToBuffer(filepath, buffer, length);
}


/* Return "~/.beebem/econet.cfg", if does not exist creates it using master
 * copy in "share/config/econet.cfg".
 */
char* GetLocation_econet_cfg(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	if (GetConfigFilePath(filepath, MAX_PATH_LEN, "econet.cfg") != TRUE)
		return NULL;

	if (CreateFileFromMaster(filepath, DATA_DIR"/config/econet.cfg")
	 == FALSE) return NULL;

	return CopyFilePathToBuffer(filepath, buffer, length);
}


/* Return "~/.beebem/phroms.cfg", if does not exist creates it using master
 * copy in "share/config/phroms.cfg".
 */
char* GetLocation_phroms_cfg(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	if (GetConfigFilePath(filepath, MAX_PATH_LEN, "phroms.cfg") != TRUE)
		return NULL;

	if (CreateFileFromMaster(filepath, DATA_DIR"/config/phroms.cfg")
	 == FALSE) return NULL;

	return CopyFilePathToBuffer(filepath, buffer, length);
}


/* Return "~/.beebem/cmos.ram", if does not exist creates it using master
 * copy in "share/state/cmos.ram".
 */
char* GetLocation_cmos_ram(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	if (GetConfigFilePath(filepath, MAX_PATH_LEN, "cmos.ram") != TRUE)
		return NULL;

	if (CreateFileFromMaster(filepath, DATA_DIR"/state/cmos.ram")
	 == FALSE) return NULL;

	return CopyFilePathToBuffer(filepath, buffer, length);
}


/* Returns "share/roms/" if a roms directory does not exist in "~/.beebem/roms".
 * Will not create a copy of the roms in the users config directory.
 */
char* GetLocation_roms(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	/* Determine users location.
	 */
	if ( GetUserConfigPath(filepath, MAX_PATH_LEN) == NULL ){
                qERROR("Unable to determine users config filepath.");
        }else{
		/* If a roms directory exists in the users config dir use that
		 * one instead of the shared one.
		 */
		if ( AppendDirToPath(filepath, "roms", MAX_PATH_LEN) ){
			if ( TestDirectory(filepath) ){
				return CopyFilePathToBuffer(filepath
				 , buffer, length);
			}
		}
	}

	return CopyFilePathToBuffer(DATA_DIR"/roms/" , buffer, length);
}


/* Return "~/.beebem/scsi", if does not exist creates it using template files
 * in "share/state/scsi/".
 */
BOOL CopyScsiImagesToUser(char *filepath)
{
	int i;
	char dst_filepath[MAX_PATH_LEN];
	char src_filepath[MAX_PATH_LEN];

#define  NUMBER_OF_FILES 9
	char files[NUMBER_OF_FILES][20]={
		"sasi0.dat",
		"scsi0.dat", "scsi0.dsc",
		"scsi1.dat", "scsi1.dsc",
		"scsi2.dat", "scsi2.dsc",
		"scsi3.dat", "scsi3.dsc"
	};

	for (i=0; i<NUMBER_OF_FILES; i++){
		strcpy(dst_filepath, filepath);
		strcat(dst_filepath, files[i]);

		/* If SCSI/SASI image does not exist, create it
		 */
		if (! TestFile(dst_filepath) ){

			strcpy(src_filepath, DATA_DIR"/media/scsi/");
			strcat(src_filepath, files[i]);
	
			/* This make be slow (especially if C copy is used) so
			 * print warning.
		 	 */
			pINFO(dL"Creating %s (may take a while).", dR
			 , files[i]);

			/* Physically copy the disk images from shared.
		 	*/
			if (CreateFileFromMaster(dst_filepath, src_filepath)
			 == FALSE){
				pERROR(dL"Failed to create '%s'.", dR
				 , dst_filepath);
			}
		}
	}

	return TRUE;
}

char* GetLocation_scsi(char *buffer, size_t length)
{
	char filepath[MAX_PATH_LEN];

	if (MakeSureBufferIsSane(buffer, length) != TRUE) return NULL;

	/* Determine users location.
	 */
	if ( GetUserConfigPath(filepath, MAX_PATH_LEN) == NULL ){
		qERROR("Unable to determine users config filepath.");
		return NULL;
	}else{
		/* Create a filepath to the users scsi disk images.
		 */
		if (! AppendDirToPath(filepath, "scsi", MAX_PATH_LEN) ){
			qERROR("Unable to determine scsi folder location!");
			return NULL;
		}else{
			/* If users scsi directory does not exist, create it.
			 */
			if (! TestDirectory(filepath) ){
				if (CreateDirectory(filepath) != TRUE){
					pERROR(dL"Unable to create '%s'."
					 , dR, filepath);
					return NULL;
				}
			}

			/* Create drive images (if required).
			 */
			if (CopyScsiImagesToUser(filepath) != TRUE){
				qERROR("Unable to create a local copy"
				 " of SCSI disk images!");
				return NULL;
			}

		}

	}
	
	return CopyFilePathToBuffer(filepath, buffer, length);
}
